# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dataset jobs."""
from urllib3.exceptions import HTTPError

from renku.core.commands.dataset import import_dataset
from renku.core.errors import ParameterError
from renku.core.management.datasets import DownloadProgressCallback
from renku.core.utils.contexts import chdir
from renku.service.jobs.constants import USER_JOB_STATE_COMPLETED, \
    USER_JOB_STATE_FAILED, USER_JOB_STATE_IN_PROGRESS
from renku.service.serializers.jobs import UserJob
from renku.service.utils import make_project_path
from renku.service.views.decorators import requires_cache


class DatasetImportJobProcess(DownloadProgressCallback):
    """Track dataset import job progress."""

    schema = UserJob()

    def __init__(self, cache, user, job):
        """Construct dataset import job progress."""
        self.cache = cache
        self.user = user
        self.job = job

        super().__init__(None, None)

    def __call__(self, description, total_size):
        """Job progress call."""
        self.job['state'] = USER_JOB_STATE_IN_PROGRESS
        self.job['extras'] = {
            'description': description,
            'total_size': total_size,
        }

        self.cache.set_job(self.user, self.job)
        super().__init__(description, total_size)

        return self

    def update(self, size):
        """Update status."""
        self.job['extras']['progress_size'] = size
        self.cache.set_job(self.user, self.job)

    def finalize(self):
        """Finalize job tracking."""
        self.job['state'] = USER_JOB_STATE_COMPLETED
        self.cache.set_job(self.user, self.job)


@requires_cache
def dataset_import(
    cache,
    user,
    user_job_id,
    project_id,
    dataset_uri,
    short_name=None,
    extract=False,
):
    """Job for dataset import."""
    user_job = cache.get_job(user, user_job_id)
    project = cache.get_project(user, project_id)

    with chdir(make_project_path(user, project)):
        try:
            import_dataset(
                dataset_uri,
                short_name,
                extract,
                progress=DatasetImportJobProcess(cache, user, user_job)
            )
        except (HTTPError, ParameterError) as exp:
            user_job['state'] = USER_JOB_STATE_FAILED
            user_job['extras']['error'] = str(exp)
            cache.set_job(user, user_job)

            # Reraise exception, so we see trace in job metadata.
            raise exp



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service dataset jobs tests."""
import os
import shutil
import uuid
from pathlib import Path

import pytest

from renku.core.commands.dataset import list_datasets
from renku.core.utils.contexts import chdir
from renku.service.jobs.constants import USER_JOB_STATE_COMPLETED, \
    USER_JOB_STATE_ENQUEUED
from renku.service.jobs.datasets import dataset_import
from renku.service.utils import make_project_path


@pytest.mark.parametrize(
    'doi', [
        '10.5281/zenodo.3239980',
        '10.5281/zenodo.3188334',
        '10.7910/DVN/TJCLKP',
    ]
)
@pytest.mark.integration
def test_dataset_import_job(doi, svc_client_cache, project):
    """Test dataset import"""
    client, cache = svc_client_cache

    user = {'user_id': 'user'}

    project_meta = {
        'project_id': uuid.uuid4().hex,
        'name': Path(project).name,
        'fullname': 'full project name',
        'email': 'my@email.com',
        'owner': 'me',
        'token': 'awesome token',
        'git_url': 'git@gitlab.com'
    }

    job_request = {
        'job_id': uuid.uuid4().hex,
        'state': USER_JOB_STATE_ENQUEUED,
    }

    cache.create_job(user, job_request)
    cache.set_project(user, project_meta['project_id'], project_meta)

    dest = make_project_path(user, project_meta)
    os.makedirs(dest.parent, exist_ok=True)
    if not (project / dest).exists():
        shutil.copytree(project, dest)

    dataset_import(
        user,
        job_request['job_id'],
        project_meta['project_id'],
        doi,
    )

    with chdir(dest):
        # TODO: #1022
        result = list_datasets(None, 'data', 'json-ld')
        assert result
        assert doi in result

    updated_job = cache.get_job(user, job_request['job_id'])

    assert USER_JOB_STATE_COMPLETED == updated_job['state']
    assert {
        'extras',
        'job_id',
        'state',
        'created_at',
        'updated_at',
    } == set(updated_job.keys())



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service project cache management."""
from renku.service.cache.base import BaseCache


class ProjectManagementCache(BaseCache):
    """Project management cache."""

    PROJECTS_SUFFIX = 'projects'

    def projects_cache_key(self, user):
        """Construct cache key based on user and projects suffix."""
        return '{0}_{1}'.format(user['user_id'], self.PROJECTS_SUFFIX)

    def set_project(self, user, project_id, metadata):
        """Cache project metadata under user hash set."""
        self.set_record(self.projects_cache_key(user), project_id, metadata)

    def get_projects(self, user):
        """Get all user cache projects."""
        return self.get_all_records(self.projects_cache_key(user))

    def get_project(self, user, project_id):
        """Get user cached project."""
        result = self.get_record(self.projects_cache_key(user), project_id)
        return result

    def invalidate_project(self, user, project_id):
        """Remove project record from hash set."""
        self.invalidate_key(self.projects_cache_key(user), project_id)

    def all_projects_iter(self):
        """Iterate over cached projects."""
        return self.scan_iter(
            '*_{0}'.format(ProjectManagementCache.PROJECTS_SUFFIX)
        )



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku JSON-LD migrations."""

from renku.core.models.migrations.dataset import migrate_absolute_paths, \
    migrate_dataset_file_id, migrate_dataset_schema, migrate_doi_identifier, \
    migrate_same_as_structure

JSONLD_MIGRATIONS = {
    'dctypes:Dataset': [migrate_dataset_schema, migrate_absolute_paths],
    'schema:Dataset': [
        migrate_absolute_paths, migrate_doi_identifier,
        migrate_same_as_structure, migrate_dataset_file_id
    ],
}



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service view decorators."""
from functools import wraps

from flask import current_app, jsonify, request
from flask_apispec import doc
from git import GitCommandError
from marshmallow import ValidationError
from redis import RedisError

from renku.core.errors import RenkuException
from renku.service.config import GIT_ACCESS_DENIED_ERROR_CODE, \
    GIT_UNKNOWN_ERROR_CODE, INTERNAL_FAILURE_ERROR_CODE, \
    INVALID_HEADERS_ERROR_CODE, INVALID_PARAMS_ERROR_CODE, \
    REDIS_EXCEPTION_ERROR_CODE, RENKU_EXCEPTION_ERROR_CODE
from renku.service.serializers.headers import UserIdentityHeaders


def requires_identity(f):
    """Wrapper which indicates that route requires user identification."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kws):
        """Represents decorated function."""
        try:
            user = UserIdentityHeaders().load(request.headers)
        except (ValidationError, KeyError):
            err_message = 'user identification is incorrect or missing'
            return jsonify(
                error={
                    'code': INVALID_HEADERS_ERROR_CODE,
                    'reason': err_message
                }
            )

        return f(user, *args, **kws)

    return decorated_function


def handle_redis_except(f):
    """Wrapper which handles Redis exceptions."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except (RedisError, OSError) as e:
            error_code = REDIS_EXCEPTION_ERROR_CODE

            return jsonify(error={
                'code': error_code,
                'reason': e.messages,
            })

    return decorated_function


@handle_redis_except
def requires_cache(f):
    """Wrapper which injects cache object into view."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        return f(current_app.config.get('cache'), *args, **kwargs)

    return decorated_function


def handle_validation_except(f):
    """Wrapper which handles marshmallow `ValidationError`."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except ValidationError as e:
            return jsonify(
                error={
                    'code': INVALID_PARAMS_ERROR_CODE,
                    'reason': e.messages,
                }
            )

    return decorated_function


def handle_renku_except(f):
    """Wrapper which handles `RenkuException`."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except RenkuException as e:
            return jsonify(
                error={
                    'code': RENKU_EXCEPTION_ERROR_CODE,
                    'reason': str(e),
                }
            )

    return decorated_function


def handle_git_except(f):
    """Wrapper which handles `RenkuException`."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except GitCommandError as e:

            error_code = GIT_ACCESS_DENIED_ERROR_CODE \
                if 'Access denied' in e.stderr else GIT_UNKNOWN_ERROR_CODE

            return jsonify(
                error={
                    'code': error_code,
                    'reason':
                        'git error: {0}'.
                        format(' '.join(e.stderr.strip().split('\n'))),
                }
            )

    return decorated_function


def accepts_json(f):
    """Wrapper which ensures only JSON payload can be in request."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        if 'Content-Type' not in request.headers:
            return jsonify(
                error={
                    'code': INVALID_HEADERS_ERROR_CODE,
                    'reason': 'invalid request headers'
                }
            )

        header_check = request.headers['Content-Type'] == 'application/json'

        if not request.is_json or not header_check:
            return jsonify(
                error={
                    'code': INVALID_HEADERS_ERROR_CODE,
                    'reason': 'invalid request payload'
                }
            )

        return f(*args, **kwargs)

    return decorated_function


def handle_base_except(f):
    """Wrapper which handles base exceptions."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except (Exception, BaseException, OSError) as e:
            error_code = INTERNAL_FAILURE_ERROR_CODE

            return jsonify(
                error={
                    'code': error_code,
                    'reason':
                        'internal error: {0}'.
                        format(' '.join(e.stderr.strip().split('\n'))),
                }
            )

    return decorated_function


def header_doc(description, tags=()):
    """Wrap additional OpenAPI header description for an endpoint."""
    return doc(
        description=description,
        params={
            'Authorization': {
                'description': (
                    'Used for users git oauth2 access. '
                    'For example: '
                    '```Authorization: Bearer asdf-qwer-zxcv```'
                ),
                'in': 'header',
                'type': 'string',
                'required': True
            },
            'Renku-User-Id': {
                'description': (
                    'Used for identification of the users. '
                    'For example: '
                    '```Renku-User-Id: sasdsa-sadsd-gsdsdh-gfdgdsd```'
                ),
                'in': 'header',
                'type': 'string',
                'required': True
            },
            'Renku-User-FullName': {
                'description': (
                    'Used for commit author signature. '
                    'For example: '
                    '```Renku-User-FullName: Rok Roskar```'
                ),
                'in': 'header',
                'type': 'string',
                'required': True
            },
            'Renku-User-Email': {
                'description': (
                    'Used for commit author signature. '
                    'For example: '
                    '```Renku-User-Email: dev@renkulab.io```'
                ),
                'in': 'header',
                'type': 'string',
                'required': True
            },
        },
        tags=list(tags),
    )



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service jobs management."""
from datetime import datetime

from renku.service.cache.base import BaseCache
from renku.service.serializers.jobs import UserJob


class JobsManagementCache(BaseCache):
    """Job management cache."""

    JOBS_SUFFIX = 'jobs'
    schema = UserJob()

    def job_cache_key(self, user):
        """Construct cache key based on user and jobs suffix."""
        return '{0}_{1}'.format(user['user_id'], self.JOBS_SUFFIX)

    def create_job(self, user, job):
        """Create new user job."""
        job['created_at'] = datetime.utcnow()
        self.set_job(user, job)

    def set_job(self, user, job):
        """Cache job state under user hash set."""
        job['updated_at'] = datetime.utcnow()
        job = self.schema.dump(job)
        self.set_record(self.job_cache_key(user), job['job_id'], job)

    def get_job(self, user, job_id):
        """Get user job."""
        record = self.get_record(self.job_cache_key(user), job_id)
        return self.schema.load(record)

    def get_jobs(self, user):
        """Get all user jobs."""
        return self.get_all_records(self.job_cache_key(user))

    def invalidate_job(self, user, job_id):
        """Remove job record from users job hash table."""
        self.invalidate_key(self.job_cache_key(user), job_id)



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Integration tests for dataset command."""
import os
import shutil
import subprocess
from collections import namedtuple
from pathlib import Path

import git
import pytest
from flaky import flaky

from renku.cli import cli
from renku.core.commands.clone import renku_clone
from renku.core.utils.contexts import chdir


@pytest.mark.parametrize(
    'doi', [{
        'doi': '10.5281/zenodo.2658634',
        'input': 'y',
        'short_name': 'pyndl_naive_discriminat_v064',
        'creator':
            'Konstantin Sering, Marc Weitz, David-Elias Künstle, '
            'Lennart Schneider',
        'version': 'v0.6.4'
    }, {
        'doi': '10.7910/DVN/F4NUMR',
        'input': 'y',
        'short_name': 'replication_data_for_ca_2',
        'creator': 'James Druckman, Martin Kifer, Michael Parkin',
        'version': '2'
    }]
)
@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_import_real_doi(runner, project, doi, sleep_after):
    """Test dataset import for existing DOI."""
    result = runner.invoke(
        cli, ['dataset', 'import', doi['doi']], input=doi['input']
    )
    assert 0 == result.exit_code
    assert 'OK' in result.output

    result = runner.invoke(cli, ['dataset', '-c', 'short_name,creators'])

    assert 0 == result.exit_code
    assert doi['short_name'] in result.output
    assert doi['creator'] in result.output

    result = runner.invoke(cli, ['dataset', 'ls-tags', doi['short_name']])
    assert 0 == result.exit_code
    assert doi['version'] in result.output


@pytest.mark.parametrize(
    'doi', [
        ('10.5281/zenodo.3239980', 'n'),
        ('10.5281/zenodo.3188334', 'y'),
        ('10.7910/DVN/TJCLKP', 'n'),
        ('10.7910/DVN/F4NUMR', 'y'),
        ('10.5281/zenodo.597964', 'y'),
        ('10.5281/zenodo.3236928', 'n'),
        ('10.5281/zenodo.2671633', 'n'),
        ('10.5281/zenodo.3237420', 'n'),
        ('10.5281/zenodo.3236928', 'n'),
        ('10.5281/zenodo.3188334', 'y'),
        ('10.5281/zenodo.3236928', 'n'),
        ('10.5281/zenodo.2669459', 'n'),
        ('10.5281/zenodo.2371189', 'n'),
        ('10.5281/zenodo.2651343', 'n'),
        ('10.5281/zenodo.1467859', 'n'),
        ('10.5281/zenodo.3240078', 'n'),
        ('10.5281/zenodo.3240053', 'n'),
        ('10.5281/zenodo.3240010', 'n'),
        ('10.5281/zenodo.3240012', 'n'),
        ('10.5281/zenodo.3240006', 'n'),
        ('10.5281/zenodo.3239996', 'n'),
        ('10.5281/zenodo.3239256', 'n'),
        ('10.5281/zenodo.3237813', 'n'),
        ('10.5281/zenodo.3239988', 'y'),
        ('10.5281/zenodo.1175627', 'y'),
        ('10.5281/zenodo.3490468', 'y'),
    ]
)
@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_import_real_param(doi, runner, project, sleep_after):
    """Test dataset import and check metadata parsing."""
    result = runner.invoke(cli, ['dataset', 'import', doi[0]], input=doi[1])

    if 'y' == doi[1]:
        assert 0 == result.exit_code
        assert 'OK' in result.output
    else:
        assert 1 == result.exit_code

    result = runner.invoke(cli, ['dataset'])
    assert 0 == result.exit_code


@pytest.mark.parametrize(
    'doi', [
        ('10.5281/zenodo.3239984', 'n'),
        ('zenodo.org/record/3239986', 'n'),
        ('10.5281/zenodo.3239982', 'n'),
    ]
)
@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_import_uri_404(doi, runner, project, sleep_after):
    """Test dataset import and check that correct exception is raised."""
    result = runner.invoke(cli, ['dataset', 'import', doi[0]], input=doi[1])
    assert 2 == result.exit_code

    result = runner.invoke(cli, ['dataset'])
    assert 0 == result.exit_code


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_import_real_doi_warnings(runner, project, sleep_after):
    """Test dataset import for existing DOI and dataset"""
    result = runner.invoke(
        cli, ['dataset', 'import', '10.5281/zenodo.1438326'], input='y'
    )
    assert 0 == result.exit_code
    assert 'Warning: Newer version found' in result.output
    assert 'OK'

    result = runner.invoke(
        cli, ['dataset', 'import', '10.5281/zenodo.1438326'], input='y'
    )
    assert 1 == result.exit_code
    assert 'Warning: Newer version found' in result.output
    assert 'Error: Dataset exists:' in result.output

    result = runner.invoke(
        cli, ['dataset', 'import', '10.5281/zenodo.597964'], input='y'
    )
    assert 0 == result.exit_code
    assert 'Warning: Newer version found' not in result.output
    assert 'Error: Dataset exists:' not in result.output
    assert 'OK' in result.output

    result = runner.invoke(cli, ['dataset'])
    assert 0 == result.exit_code
    assert 'pyndl_naive_discriminat' in result.output


@pytest.mark.parametrize(
    'doi', [('10.5281/zenodo.5979642342', 'Zenodo'),
            ('10.7910/DVN/S8MSVFXXXX', 'DVN')]
)
@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_import_fake_doi(runner, project, doi):
    """Test error raising for non-existing DOI."""
    result = runner.invoke(cli, ['dataset', 'import', doi[0]], input='y')

    assert 2 == result.exit_code
    assert 'URI not found.' in result.output \
           or 'Provider {} not found'.format(doi[1]) in result.output


@pytest.mark.parametrize(
    'url', [
        'https://zenodo.org/record/2621208',
        (
            'https://dataverse.harvard.edu/dataset.xhtml'
            '?persistentId=doi:10.7910/DVN/F4NUMR'
        )
    ]
)
@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_import_real_http(runner, project, url, sleep_after):
    """Test dataset import through HTTPS."""
    result = runner.invoke(cli, ['dataset', 'import', url], input='y')

    assert 0 == result.exit_code
    assert 'OK' in result.output


@pytest.mark.parametrize(
    'url', [
        'https://zenodo.org/record/2621201248',
        'https://dataverse.harvard.edu/dataset.xhtml' +
        '?persistentId=doi:10.7910/DVN/F4NUMRXXXX'
    ]
)
@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_import_fake_http(runner, project, url):
    """Test dataset import through HTTPS."""
    result = runner.invoke(cli, ['dataset', 'import', url], input='y')

    assert 2 == result.exit_code
    assert 'URI not found.' in result.output


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_import_and_extract(runner, project, client, sleep_after):
    """Test dataset import and extract files."""
    url = 'https://zenodo.org/record/2658634'
    result = runner.invoke(
        cli, ['dataset', 'import', '--extract', '--short-name', 'remote', url],
        input='y'
    )
    assert 0 == result.exit_code

    with client.with_dataset('remote') as dataset:
        extracted_file = 'data/remote/quantling-pyndl-c34259c/doc/make.bat'
        assert dataset.find_file(extracted_file)


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_import_different_names(runner, client, sleep_after):
    """Test can import same DOI under different names."""
    doi = '10.5281/zenodo.2658634'
    result = runner.invoke(
        cli, ['dataset', 'import', '--short-name', 'name-1', doi], input='y'
    )
    assert 0 == result.exit_code
    assert 'OK' in result.output

    result = runner.invoke(
        cli, ['dataset', 'import', '--short-name', 'name-2', doi], input='y'
    )
    assert 0 == result.exit_code
    assert 'OK' in result.output


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_import_ignore_uncompressed_files(
    runner, project, sleep_after
):
    """Test dataset import ignores uncompressed files."""
    url = 'https://zenodo.org/record/3251128'
    result = runner.invoke(
        cli, ['dataset', 'import', '--extract', url], input='y'
    )
    assert 0 == result.exit_code
    assert 'Gorne_Diaz_database_2019.csv' in result.output


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_reimport_removed_dataset(runner, project, sleep_after):
    """Test re-importing of deleted datasets works."""
    doi = '10.5281/zenodo.2658634'
    result = runner.invoke(
        cli, ['dataset', 'import', doi, '--short-name', 'my-dataset'],
        input='y'
    )
    assert 0 == result.exit_code

    result = runner.invoke(cli, ['dataset', 'rm', 'my-dataset'])
    assert 0 == result.exit_code

    result = runner.invoke(
        cli, ['dataset', 'import', doi, '--short-name', 'my-dataset'],
        input='y'
    )
    assert 0 == result.exit_code


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_import_preserve_names(runner, project, sleep_after):
    """Test import keeps original file names."""
    doi = '10.7910/DVN/F4NUMR'
    result = runner.invoke(cli, ['dataset', 'import', doi], input='y')
    assert 0 == result.exit_code

    result = runner.invoke(cli, ['dataset', 'ls-files'])
    assert 0 == result.exit_code
    assert 'Data Key 2002-2006' in result.output


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.parametrize(
    'provider,params,output',
    [('zenodo', [], 'zenodo.org/deposit'),
     ('dataverse', ['--dataverse-name', 'SDSC-Test'], 'doi:')]
)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_export_upload_file(
    runner, project, tmpdir, client, zenodo_sandbox, dataverse_demo, provider,
    params, output
):
    """Test successful uploading of a file to Zenodo deposit."""
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])

    assert 0 == result.exit_code
    assert 'OK' in result.output

    # create data file
    new_file = tmpdir.join('datafile.csv')
    new_file.write('1,2,3')

    # add data to dataset
    result = runner.invoke(
        cli, ['dataset', 'add', 'my-dataset',
              str(new_file)]
    )
    assert 0 == result.exit_code

    with client.with_dataset('my-dataset') as dataset:
        dataset.description = 'awesome dataset'
        dataset.creator[0].affiliation = 'eth'

    data_repo = git.Repo(str(project))
    data_repo.git.add(update=True)
    data_repo.index.commit('metadata updated')

    result = runner.invoke(
        cli, ['dataset', 'export', 'my-dataset', provider] + params
    )

    assert 0 == result.exit_code
    assert 'Exported to:' in result.output
    assert output in result.output


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.parametrize(
    'provider,params,output',
    [('zenodo', [], 'zenodo.org/deposit'),
     ('dataverse', ['--dataverse-name', 'SDSC-Test'], 'doi:')]
)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_export_upload_tag(
    runner, project, tmpdir, client, zenodo_sandbox, dataverse_demo, provider,
    params, output
):
    """Test successful uploading of a file to Zenodo deposit."""
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    # create data file
    new_file = tmpdir.join('datafile.csv')
    new_file.write('1,2,3')

    # add data to dataset
    result = runner.invoke(
        cli, ['dataset', 'add', 'my-dataset',
              str(new_file)]
    )
    assert 0 == result.exit_code

    with client.with_dataset('my-dataset') as dataset:
        dataset.description = 'awesome dataset'
        dataset.creator[0].affiliation = 'eth'

    data_repo = git.Repo(str(project))
    data_repo.git.add(update=True)
    data_repo.index.commit('metadata updated')

    # tag dataset
    result = runner.invoke(cli, ['dataset', 'tag', 'my-dataset', '1.0'])
    assert 0 == result.exit_code

    # create data file
    new_file = tmpdir.join('datafile2.csv')
    new_file.write('1,2,3,4')

    # add data to dataset
    result = runner.invoke(
        cli, ['dataset', 'add', 'my-dataset',
              str(new_file)]
    )
    assert 0 == result.exit_code

    # tag dataset
    result = runner.invoke(cli, ['dataset', 'tag', 'my-dataset', '2.0'])
    assert 0 == result.exit_code

    result = runner.invoke(
        cli, ['dataset', 'export', 'my-dataset', provider] + params, input='3'
    )

    assert 0 == result.exit_code
    assert 'Exported to:' in result.output
    assert output in result.output
    assert '2/2' in result.output

    result = runner.invoke(
        cli, ['dataset', 'export', 'my-dataset', provider] + params, input='2'
    )

    assert 0 == result.exit_code
    assert 'Exported to:' in result.output
    assert output in result.output
    assert '1/1' in result.output

    result = runner.invoke(
        cli, ['dataset', 'export', 'my-dataset', provider] + params, input='1'
    )

    assert 0 == result.exit_code
    assert 'Exported to:' in result.output
    assert output in result.output
    assert '2/2' in result.output


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.parametrize(
    'provider,params,output',
    [('zenodo', [], 'zenodo.org/deposit'),
     ('dataverse', ['--dataverse-name', 'SDSC-Test'], 'doi:')]
)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_export_upload_multiple(
    runner, project, tmpdir, client, zenodo_sandbox, dataverse_demo, provider,
    params, output
):
    """Test successful uploading of a files to Zenodo deposit."""
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])

    assert 0 == result.exit_code
    assert 'OK' in result.output

    # create data file
    paths = []
    for i in range(3):
        new_file = tmpdir.join('file_{0}'.format(i))
        new_file.write(str(i))
        paths.append(str(new_file))

    # add data
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'my-dataset'] + paths,
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    with client.with_dataset('my-dataset') as dataset:
        dataset.description = 'awesome dataset'
        dataset.creator[0].affiliation = 'eth'

    data_repo = git.Repo(str(project))
    data_repo.git.add(update=True)
    data_repo.index.commit('metadata updated')

    result = runner.invoke(
        cli, ['dataset', 'export', 'my-dataset', provider] + params
    )

    assert 0 == result.exit_code
    assert 'Exported to:' in result.output
    assert output in result.output


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_export_upload_failure(runner, tmpdir, client, zenodo_sandbox):
    """Test failed uploading of a file to Zenodo deposit."""
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])

    assert 0 == result.exit_code
    assert 'OK' in result.output

    # create data file
    new_file = tmpdir.join('datafile.csv')
    new_file.write('1,2,3')

    # add data to dataset
    result = runner.invoke(
        cli, ['dataset', 'add', 'my-dataset',
              str(new_file)]
    )
    assert 0 == result.exit_code

    result = runner.invoke(cli, ['dataset', 'export', 'my-dataset', 'zenodo'])

    assert 1 == result.exit_code
    assert 'metadata.creators.0.affiliation' in result.output
    assert 'metadata.description' in result.output


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.parametrize(
    'provider,params,output',
    [('zenodo', [], 'zenodo.org/record'),
     ('dataverse', ['--dataverse-name', 'SDSC-Test'], 'doi:')]
)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_export_published_url(
    runner, project, tmpdir, client, zenodo_sandbox, dataverse_demo, provider,
    params, output
):
    """Test publishing of dataset."""
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])

    assert 0 == result.exit_code
    assert 'OK' in result.output

    # create data file
    new_file = tmpdir.join('datafile.csv')
    new_file.write('1,2,3')

    # add data to dataset
    result = runner.invoke(
        cli, ['dataset', 'add', 'my-dataset',
              str(new_file)]
    )
    assert 0 == result.exit_code

    with client.with_dataset('my-dataset') as dataset:
        dataset.description = 'awesome dataset'
        dataset.creator[0].affiliation = 'eth'

    data_repo = git.Repo(str(project))
    data_repo.git.add(update=True)
    data_repo.index.commit('metadata updated')

    result = runner.invoke(
        cli,
        ['dataset', 'export', 'my-dataset', provider, '--publish'] + params
    )

    assert 0 == result.exit_code
    assert 'Exported to:' in result.output
    assert output in result.output


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_export_dataset_wrong_provider(
    runner, project, tmpdir, client, zenodo_sandbox
):
    """Test non-existing provider."""
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])

    assert 0 == result.exit_code
    assert 'OK' in result.output

    # create data file
    new_file = tmpdir.join('datafile.csv')
    new_file.write('1,2,3')

    # add data to dataset
    result = runner.invoke(
        cli, ['dataset', 'add', 'my-dataset',
              str(new_file)]
    )
    assert 0 == result.exit_code

    result = runner.invoke(
        cli, ['dataset', 'export', 'my-dataset', 'unsupported-provider']
    )
    assert 2 == result.exit_code
    assert 'Unknown provider.' in result.output


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_export(runner, client, project):
    """Check dataset not found exception raised."""
    result = runner.invoke(
        cli, ['dataset', 'export', 'doesnotexists', 'somewhere']
    )

    assert 2 == result.exit_code
    assert 'Dataset is not found.' in result.output


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.parametrize(
    'provider,params', [('zenodo', []),
                        ('dataverse', ['--dataverse-name', 'SDSC-Test'])]
)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_export_dataset_unauthorized(
    runner, project, client, tmpdir, zenodo_sandbox, dataverse_demo, provider,
    params
):
    """Test unauthorized exception raised."""
    client.set_value(provider, 'access_token', 'not-a-token')
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    # create data file
    new_file = tmpdir.join('datafile.csv')
    new_file.write('1,2,3')

    # add data to dataset
    result = runner.invoke(
        cli, ['dataset', 'add', 'my-dataset',
              str(new_file)]
    )
    assert 0 == result.exit_code

    result = runner.invoke(
        cli, ['dataset', 'export', 'my-dataset', provider] + params
    )

    assert 1 == result.exit_code
    assert 'Access unauthorized - update access token.' in result.output

    secret = client.get_value('zenodo', 'secret')
    assert secret is None


@pytest.mark.integration
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_export_dataverse_no_dataverse_name(
    runner, project, client, dataverse_demo
):
    """Test export without providing a dataverse name."""
    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code

    result = runner.invoke(
        cli, ['dataset', 'export', 'my-dataset', 'dataverse']
    )

    assert 2 == result.exit_code
    assert 'Dataverse name is required.' in result.output


@pytest.mark.integration
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_export_dataverse_no_dataverse_url(runner, client, dataverse_demo):
    """Test export without providing a dataverse server url."""
    client.remove_value('dataverse', 'server_url')
    client.repo.git.add('.renku/renku.ini')
    client.repo.index.commit('renku.ini')

    result = runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code

    result = runner.invoke(
        cli, [
            'dataset', 'export', 'my-dataset', 'dataverse', '--dataverse-name',
            'name'
        ]
    )

    assert 2 == result.exit_code
    assert 'Dataverse server URL is required.' in result.output


@pytest.mark.integration
@pytest.mark.parametrize(
    'params,path',
    [
        # add data with no destination
        (['-s', 'docker'], 'data/remote/docker/r/Dockerfile'),
        (['-s', 'docker/r/Dockerfile'], 'data/remote/Dockerfile'),
        # add data to a non-existing destination
        (['-s', 'docker', '-d', 'new'], 'data/remote/new/r/Dockerfile'),
        (['-s', 'docker/r', '-d', 'new'], 'data/remote/new/Dockerfile'),
        (['-s', 'docker/r/Dockerfile', '-d', 'new'], 'data/remote/new'),
        # add data to an existing destination
        (['-s', 'docker', '-d', 'existing'
          ], 'data/remote/existing/docker/r/Dockerfile'),
        (['-s', 'docker/r', '-d', 'existing'
          ], 'data/remote/existing/r/Dockerfile'),
        (['-s', 'docker/r/Dockerfile', '-d', 'existing'
          ], 'data/remote/existing/Dockerfile'),
    ]
)
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_add_data_from_git(runner, client, params, path):
    """Test add data to datasets from a git repository."""
    REMOTE = 'https://github.com/SwissDataScienceCenter/renku-jupyter.git'

    # create a dataset and add a file to it
    result = runner.invoke(
        cli,
        [
            'dataset', 'add', 'remote', '--create', '--ref', '0.3.0', '-s',
            'LICENSE', '-d', 'existing/LICENSE', REMOTE
        ],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    result = runner.invoke(
        cli,
        ['dataset', 'add', 'remote', '--ref', '0.3.0', REMOTE] + params,
        catch_exceptions=False,
    )

    assert 0 == result.exit_code
    assert Path(path).exists()


@pytest.mark.integration
@pytest.mark.skip(reason='temporarily disabled, until #1035')
@flaky(max_runs=10, min_passes=1)
def test_add_from_git_copies_metadata(runner, client):
    """Test an import from a git repository keeps creators name."""
    # create a dataset and add a file to it
    result = runner.invoke(
        cli,
        [
            'dataset', 'add', 'remote', '--create', '--ref', 'v0.3.0', '-s',
            'README.rst',
            'https://github.com/SwissDataScienceCenter/renku-python.git'
        ],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    dataset = client.load_dataset('remote')
    assert dataset.files[0].name == 'README.rst'
    assert 'mailto:jiri.kuncar@gmail.com' in str(dataset.files[0].creator)
    assert 'mailto:rokroskar@gmail.co' in str(dataset.files[0].creator)


@pytest.mark.integration
@pytest.mark.parametrize(
    'params,n_urls,message', [
        ([], 0, 'No URL is specified'),
        (['-s', 'file', '-d', 'new-file'], 0, 'No URL is specified'),
        (['-s', 'file'], 2, 'Cannot add multiple URLs'),
        (['-d', 'file'], 2, 'Cannot add multiple URLs'),
        (['-s', 'non-existing'], 1, 'No such file or directory'),
        (['-s', 'docker', '-d', 'LICENSE'
          ], 1, 'Cannot copy multiple files or directories to a file'),
        (['-s', 'LICENSE', '-s', 'Makefile', '-d', 'LICENSE'
          ], 1, 'Cannot copy multiple files or directories to a file'),
        (['-d', 'LICENSE'], 1, 'Cannot copy repo to file'),
    ]
)
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_usage_error_in_add_from_git(runner, client, params, n_urls, message):
    """Test user's errors when adding to a dataset from a git repository."""
    remote = 'https://github.com/SwissDataScienceCenter/renku-jupyter.git'

    # create a dataset and add a file to it
    result = runner.invoke(
        cli,
        [
            'dataset', 'add', 'remote', '--create', '--ref', '0.3.0', '-s',
            'LICENSE', remote
        ],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    urls = n_urls * [remote]

    result = runner.invoke(
        cli,
        ['dataset', 'add', 'remote', '--ref', '0.3.0'] + params + urls,
        catch_exceptions=False,
    )
    assert 2 == result.exit_code
    assert message in result.output


def read_dataset_file_metadata(client, dataset_name, filename):
    """Return metadata from dataset's YAML file."""
    with client.with_dataset(dataset_name) as dataset:
        assert client.get_dataset_path(dataset.name).exists()

        for file_ in dataset.files:
            if file_.path.endswith(filename):
                return file_


@pytest.mark.integration
@pytest.mark.parametrize(
    'params', [[], ['-I', 'CHANGES.rst'], ['-I', 'C*'], ['remote']]
)
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_update(client, runner, params):
    """Test local copy is updated when remote file is updates."""
    # Add dataset to project
    result = runner.invoke(
        cli, [
            'dataset', 'add', '--create', 'remote', '--ref', 'v0.3.0', '-s',
            'CHANGES.rst',
            'https://github.com/SwissDataScienceCenter/renku-python.git'
        ],
        catch_exceptions=False
    )
    assert 0 == result.exit_code

    before = read_dataset_file_metadata(client, 'remote', 'CHANGES.rst')

    result = runner.invoke(
        cli, ['dataset', 'update'] + params, catch_exceptions=False
    )
    assert 0 == result.exit_code

    after = read_dataset_file_metadata(client, 'remote', 'CHANGES.rst')
    assert after._id == before._id
    assert after._label != before._label
    assert after.added == before.added
    assert after.url == before.url
    assert after.based_on._id == before.based_on._id
    assert after.based_on._label != before.based_on._label
    assert after.based_on.path == before.based_on.path
    assert after.based_on.based_on is None


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_update_remove_file(client, runner):
    """Test local copy is removed when remote file is removed."""
    # Add dataset to project
    result = runner.invoke(
        cli, [
            'dataset', 'add', '--create', 'remote', '-s', 'docs/authors.rst',
            '--ref', 'v0.3.0',
            'https://github.com/SwissDataScienceCenter/renku-python.git'
        ],
        catch_exceptions=False
    )
    assert 0 == result.exit_code
    file_path = client.path / 'data' / 'remote' / 'authors.rst'
    assert file_path.exists()

    # docs/authors.rst does not exists in v0.5.0

    result = runner.invoke(
        cli, ['dataset', 'update', '--ref', 'v0.5.0'], catch_exceptions=False
    )
    assert 0 == result.exit_code
    assert 'Some files are deleted from remote.' in result.output
    assert file_path.exists()

    result = runner.invoke(
        cli, ['dataset', 'update', '--ref', 'v0.5.0', '--delete'],
        catch_exceptions=False
    )
    assert 0 == result.exit_code
    assert not file_path.exists()


@pytest.mark.integration
@pytest.mark.parametrize(
    'params', [['-I', 'non-existing'], ['non-existing-dataset']]
)
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_invalid_update(client, runner, params):
    """Test updating a non-existing path."""
    # Add dataset to project
    result = runner.invoke(
        cli, [
            'dataset', 'add', '--create', 'remote', '-s', 'docs/authors.rst',
            '--ref', 'v0.3.0',
            'https://github.com/SwissDataScienceCenter/renku-python.git'
        ],
        catch_exceptions=False
    )
    assert 0 == result.exit_code

    result = runner.invoke(
        cli, ['dataset', 'update'] + params, catch_exceptions=False
    )
    assert 2 == result.exit_code


@pytest.mark.integration
@pytest.mark.parametrize(
    'params',
    [[], ['-I', 'CHANGES.rst'], ['-I', 'CH*'], ['dataset-1', 'dataset-2']]
)
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_dataset_update_multiple_datasets(
    client, runner, data_repository, directory_tree, params
):
    """Test update with multiple datasets."""
    path1 = client.path / 'data' / 'dataset-1' / 'CHANGES.rst'
    path2 = client.path / 'data' / 'dataset-2' / 'CHANGES.rst'
    # Add dataset to project
    result = runner.invoke(
        cli, [
            'dataset', 'add', '--create', 'dataset-1', '--ref', 'v0.3.0', '-s',
            'CHANGES.rst',
            'https://github.com/SwissDataScienceCenter/renku-python.git'
        ],
        catch_exceptions=False
    )
    assert 0 == result.exit_code
    result = runner.invoke(
        cli, [
            'dataset', 'add', '--create', 'dataset-2', '--ref', 'v0.3.0', '-s',
            'CHANGES.rst',
            'https://github.com/SwissDataScienceCenter/renku-python.git'
        ],
        catch_exceptions=False
    )
    assert 0 == result.exit_code

    assert 'v0.4.0' not in path1.read_text()
    assert 'v0.4.0' not in path2.read_text()

    result = runner.invoke(
        cli, ['dataset', 'update'] + params, catch_exceptions=False
    )
    assert 0 == result.exit_code

    assert 'v0.4.0' in path1.read_text()
    assert 'v0.4.0' in path2.read_text()


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_empty_update(client, runner, data_repository, directory_tree):
    """Test update when nothing changed does not create a commit."""
    # Add dataset to project
    result = runner.invoke(
        cli, [
            'dataset', 'add', '--create', 'remote', '--ref', 'v0.3.0', '-s',
            'CHANGES.rst',
            'https://github.com/SwissDataScienceCenter/renku-python.git'
        ],
        catch_exceptions=False
    )
    assert 0 == result.exit_code

    commit_sha_before = client.repo.head.object.hexsha
    result = runner.invoke(
        cli, ['dataset', 'update', '--ref', 'v0.3.0'], catch_exceptions=False
    )
    assert 0 == result.exit_code
    commit_sha_after = client.repo.head.object.hexsha
    assert commit_sha_after == commit_sha_before


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_import_from_renku_project(tmpdir, client, runner):
    """Test an imported dataset from other renku repos will have metadata."""
    from renku.core.management import LocalClient

    remote = 'https://dev.renku.ch/gitlab/virginiafriedrich/datasets-test.git'

    path = tmpdir.strpath
    os.environ['GIT_LFS_SKIP_SMUDGE'] = '1'
    git.Repo.clone_from(remote, path, recursive=True)

    remote_client = LocalClient(path)
    file_ = read_dataset_file_metadata(
        remote_client, 'zhbikes',
        '2019_verkehrszaehlungen_werte_fussgaenger_velo.csv'
    )

    result = runner.invoke(
        cli,
        [
            'dataset', 'add', '--create', 'remote-dataset', '-s',
            'data/zhbikes/2019_verkehrszaehlungen_werte_fussgaenger_velo.csv',
            '-d', 'file', '--ref', 'b973db5', remote
        ],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    metadata = read_dataset_file_metadata(client, 'remote-dataset', 'file')
    assert metadata.creator[0].name == file_.creator[0].name
    assert metadata.based_on._id == file_._id
    assert metadata.based_on._label == file_._label
    assert metadata.based_on.path == file_.path
    assert metadata.based_on.based_on is None
    assert metadata.based_on.url == remote


@pytest.mark.integration
@pytest.mark.parametrize(
    'ref', ['v0.3.0', 'fe6ec65cc84bcf01e879ef38c0793208f7fab4bb']
)
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_add_specific_refs(ref, runner, client):
    """Test adding a specific version of files."""
    FILENAME = 'CHANGES.rst'
    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'dataset'])
    assert 0 == result.exit_code

    # add data from a git repo
    result = runner.invoke(
        cli, [
            'dataset', 'add', 'dataset', '-s', FILENAME, '--ref', ref,
            'https://github.com/SwissDataScienceCenter/renku-python.git'
        ]
    )
    assert 0 == result.exit_code
    content = (client.path / 'data' / 'dataset' / FILENAME).read_text()
    assert 'v0.3.0' in content
    assert 'v0.3.1' not in content


@pytest.mark.integration
@pytest.mark.parametrize(
    'ref', ['v0.3.1', '27e29abd409c83129a3fdb8b8b0b898b23bcb229']
)
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_update_specific_refs(ref, runner, client):
    """Test updating to a specific version of files."""
    filename = 'CHANGES.rst'
    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'dataset'])
    assert 0 == result.exit_code

    # add data from a git repo
    result = runner.invoke(
        cli, [
            'dataset', 'add', 'dataset', '-s', filename, '--ref', 'v0.3.0',
            'https://github.com/SwissDataScienceCenter/renku-python.git'
        ]
    )
    assert 0 == result.exit_code
    content = (client.path / 'data' / 'dataset' / filename).read_text()
    assert 'v0.3.1' not in content

    # update data to a later version
    result = runner.invoke(cli, ['dataset', 'update', '--ref', ref])
    assert 0 == result.exit_code
    content = (client.path / 'data' / 'dataset' / filename).read_text()
    assert 'v0.3.1' in content
    assert 'v0.3.2' not in content


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_update_with_multiple_remotes_and_ref(runner, client):
    """Test updating fails when ref is ambiguous."""
    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'dataset'])
    assert 0 == result.exit_code

    # add data from a git repo
    result = runner.invoke(
        cli, [
            'dataset', 'add', 'dataset', '-s', 'CHANGES.rst',
            'https://github.com/SwissDataScienceCenter/renku-python.git'
        ]
    )
    assert 0 == result.exit_code

    # add data from another git repo
    result = runner.invoke(
        cli, [
            'dataset', 'add', 'dataset', '-s', 'LICENSE',
            'https://github.com/SwissDataScienceCenter/renku-notebooks.git'
        ]
    )
    assert 0 == result.exit_code

    # update data to a later version
    result = runner.invoke(cli, ['dataset', 'update', '--ref', 'any-value'])
    assert 2 == result.exit_code
    assert 'Cannot use "--ref" with more than one Git repo' in result.output


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_files_are_tracked_in_lfs(runner, client):
    """Test files added from a Git repo are tacked in Git LFS."""
    filename = 'CHANGES.rst'
    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'dataset'])
    assert 0 == result.exit_code

    # add data from a git repo
    result = runner.invoke(
        cli, [
            'dataset', 'add', 'dataset', '-s', filename,
            'https://github.com/SwissDataScienceCenter/renku-python.git'
        ]
    )
    assert 0 == result.exit_code
    path = 'data/dataset/{}'.format(filename)
    assert path in subprocess.check_output(['git', 'lfs', 'ls-files']).decode()


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_renku_clone(runner, monkeypatch):
    """Test cloning of a Renku repo and existence of required settings."""
    from renku.core.management.storage import StorageApiMixin

    remote = 'https://dev.renku.ch/gitlab/virginiafriedrich/datasets-test.git'

    with runner.isolated_filesystem() as project_path:
        result = runner.invoke(cli, ['clone', remote, project_path])
        assert 0 == result.exit_code
        assert (Path(project_path) / 'Dockerfile').exists()

        # Check Git hooks are installed
        result = runner.invoke(cli, ['githooks', 'install'])
        assert 0 == result.exit_code
        assert 'Hook already exists.' in result.output

        # Check Git LFS is enabled
        with monkeypatch.context() as monkey:
            # Pretend that git-lfs is not installed.
            monkey.setattr(StorageApiMixin, 'storage_installed', False)
            # Repo is using external storage but it's not installed.
            result = runner.invoke(cli, ['run', 'touch', 'output'])
            assert 'is not configured' in result.output
            assert 1 == result.exit_code


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
@pytest.mark.skip(reason='temporarily disabled, until #1035')
def test_renku_clone_with_config(tmpdir):
    """Test cloning of a Renku repo and existence of required settings."""
    remote = 'https://dev.renku.ch/gitlab/virginiafriedrich/datasets-test.git'

    with chdir(str(tmpdir)):
        renku_clone(
            remote,
            config={
                'user.name': 'sam',
                'user.email': 's@m.i',
                'filter.lfs.custom': '0'
            }
        )

        repo = git.Repo('datasets-test')
        reader = repo.config_reader()
        reader.values()

        lfs_config = dict(reader.items('filter.lfs'))
        assert '0' == lfs_config.get('custom')


@pytest.mark.integration
@pytest.mark.parametrize(
    'path,expected_path', [('', 'datasets-test'), ('.', '.'),
                           ('new-name', 'new-name')]
)
@flaky(max_runs=10, min_passes=1)
def test_renku_clone_uses_project_name(
    runner, monkeypatch, path, expected_path
):
    """Test renku clone uses project name as target-path by default."""
    remote = 'https://dev.renku.ch/gitlab/virginiafriedrich/datasets-test.git'

    with runner.isolated_filesystem() as project_path:
        result = runner.invoke(cli, ['clone', remote, path])
        assert 0 == result.exit_code
        assert (Path(project_path) / expected_path / 'Dockerfile').exists()


@pytest.mark.integration
@flaky(max_runs=10, min_passes=1)
def test_add_removes_credentials(runner, client):
    """Check removal of credentials during adding of remote data files."""
    url = 'https://username:password@example.com/index.html'
    result = runner.invoke(cli, ['dataset', 'add', '-c', 'my-dataset', url])
    assert 0 == result.exit_code

    with client.with_dataset('my-dataset') as dataset:
        file_ = dataset.files[0]
        assert file_.url == 'https://example.com/index.html'


@pytest.mark.integration
def test_check_disk_space(runner, client, monkeypatch):
    """Check adding to dataset prompts if disk space is not enough."""
    url = 'https://example.com/index.html'

    def disk_usage(_):
        """Mocked response."""
        Usage = namedtuple('Usage', 'free')
        return Usage(free=0)

    monkeypatch.setattr(shutil, 'disk_usage', disk_usage)

    result = runner.invoke(
        cli,
        ['dataset', 'add', '-c', 'my-data', url],
        catch_exceptions=False,
    )
    assert 1 == result.exit_code
    assert 'Insufficient disk space' in result.output

    result = runner.invoke(cli, ['dataset', 'ls-files'])
    assert 'index.html' not in result.output



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache views."""
import os
import shutil
from pathlib import Path

import patoolib
from flask import Blueprint, jsonify, request
from flask_apispec import marshal_with, use_kwargs
from marshmallow import EXCLUDE
from patoolib.util import PatoolError

from renku.core.commands.clone import renku_clone
from renku.service.config import CACHE_UPLOADS_PATH, \
    INVALID_PARAMS_ERROR_CODE, SERVICE_PREFIX, SUPPORTED_ARCHIVES
from renku.service.serializers.cache import FileListResponse, \
    FileListResponseRPC, FileUploadContext, FileUploadRequest, \
    FileUploadResponse, FileUploadResponseRPC, ProjectCloneContext, \
    ProjectCloneRequest, ProjectCloneResponse, ProjectCloneResponseRPC, \
    ProjectListResponse, ProjectListResponseRPC, extract_file
from renku.service.utils import make_project_path, valid_file
from renku.service.views.decorators import accepts_json, handle_base_except, \
    handle_git_except, handle_renku_except, handle_validation_except, \
    header_doc, requires_cache, requires_identity

CACHE_BLUEPRINT_TAG = 'cache'
cache_blueprint = Blueprint('cache', __name__, url_prefix=SERVICE_PREFIX)


@marshal_with(FileListResponseRPC)
@header_doc(description='List uploaded files.', tags=(CACHE_BLUEPRINT_TAG, ))
@cache_blueprint.route(
    '/cache.files_list',
    methods=['GET'],
    provide_automatic_options=False,
)
@handle_base_except
@handle_git_except
@handle_renku_except
@handle_validation_except
@requires_cache
@requires_identity
def list_uploaded_files_view(user, cache):
    """List uploaded files ready to be added to projects."""
    files = filter(None, [valid_file(user, f) for f in cache.get_files(user)])

    response_payload = {
        'files':
            sorted(
                files, key=lambda rec: (rec['is_dir'], rec['relative_path'])
            )
    }

    response = FileListResponseRPC().load({
        'result': FileListResponse().load(response_payload)
    })
    return jsonify(response)


@use_kwargs(FileUploadRequest)
@marshal_with(FileUploadResponseRPC)
@header_doc(
    description='Upload file or archive of files.',
    tags=(CACHE_BLUEPRINT_TAG, ),
)
@cache_blueprint.route(
    '/cache.files_upload',
    methods=['POST'],
    provide_automatic_options=False,
)
@handle_base_except
@handle_git_except
@handle_renku_except
@handle_validation_except
@requires_cache
@requires_identity
def upload_file_view(user, cache):
    """Upload file or archive of files."""
    file = extract_file(request)

    response_builder = {
        'file_name': file.filename,
        'content_type': file.content_type,
        'is_archive': file.content_type in SUPPORTED_ARCHIVES
    }
    response_builder.update(FileUploadRequest().load(request.args))

    user_cache_dir = CACHE_UPLOADS_PATH / user['user_id']
    user_cache_dir.mkdir(exist_ok=True)

    file_path = user_cache_dir / file.filename
    if file_path.exists():
        if response_builder.get('override_existing', False):
            file_path.unlink()
        else:
            return jsonify(
                error={
                    'code': INVALID_PARAMS_ERROR_CODE,
                    'reason': 'file exists',
                }
            )

    file.save(str(file_path))

    files = []
    if response_builder['unpack_archive'] and response_builder['is_archive']:
        unpack_dir = '{0}.unpacked'.format(file_path.name)
        temp_dir = file_path.parent / Path(unpack_dir)
        if temp_dir.exists():
            shutil.rmtree(str(temp_dir))
        temp_dir.mkdir(exist_ok=True)

        try:
            patoolib.extract_archive(str(file_path), outdir=str(temp_dir))
        except PatoolError:
            return jsonify(
                error={
                    'code': INVALID_PARAMS_ERROR_CODE,
                    'reason': 'unable to unpack archive'
                }
            )

        for file_ in temp_dir.glob('**/*'):
            relative_path = file_.relative_to(
                CACHE_UPLOADS_PATH / user['user_id']
            )

            file_obj = {
                'file_name': file_.name,
                'file_size': os.stat(str(file_path)).st_size,
                'relative_path': str(relative_path),
                'is_dir': relative_path.is_dir(),
            }

            files.append(FileUploadContext().load(file_obj, unknown=EXCLUDE))

    else:
        relative_path = file_path.relative_to(
            CACHE_UPLOADS_PATH / user['user_id']
        )
        response_builder['file_size'] = os.stat(str(file_path)).st_size
        response_builder['relative_path'] = str(relative_path)
        response_builder['is_dir'] = relative_path.is_dir()

        files.append(
            FileUploadContext().load(response_builder, unknown=EXCLUDE)
        )

    response = FileUploadResponseRPC().load({
        'result': FileUploadResponse().load({'files': files})
    })
    cache.set_files(user, files)

    return jsonify(response)


@use_kwargs(ProjectCloneRequest)
@marshal_with(ProjectCloneResponseRPC)
@header_doc(
    'Clone a remote project. If the project is cached already, '
    'new clone operation will override the old cache state.',
    tags=(CACHE_BLUEPRINT_TAG, )
)
@cache_blueprint.route(
    '/cache.project_clone',
    methods=['POST'],
    provide_automatic_options=False,
)
@handle_base_except
@handle_git_except
@handle_renku_except
@handle_validation_except
@requires_cache
@requires_identity
@accepts_json
def project_clone(user, cache):
    """Clone a remote repository."""
    ctx = ProjectCloneContext().load(
        (lambda a, b: a.update(b) or a)(request.json, user),
        unknown=EXCLUDE,
    )

    local_path = make_project_path(user, ctx)

    if local_path.exists():
        shutil.rmtree(str(local_path))

        for project in cache.get_projects(user):
            if project['git_url'] == ctx['git_url']:
                cache.invalidate_project(user, project['project_id'])

    local_path.mkdir(parents=True, exist_ok=True)
    renku_clone(
        ctx['url_with_auth'],
        local_path,
        depth=ctx['depth'],
        raise_git_except=True,
        config={
            'user.name': ctx['fullname'],
            'user.email': ctx['email'],
        }
    )
    cache.set_project(user, ctx['project_id'], ctx)

    response = ProjectCloneResponseRPC().load({
        'result': ProjectCloneResponse().load(ctx, unknown=EXCLUDE)
    })
    return jsonify(response)


@marshal_with(ProjectListResponseRPC)
@header_doc(
    'List cached projects.',
    tags=(CACHE_BLUEPRINT_TAG, ),
)
@cache_blueprint.route(
    '/cache.project_list',
    methods=['GET'],
    provide_automatic_options=False,
)
@handle_base_except
@handle_git_except
@handle_renku_except
@handle_validation_except
@requires_cache
@requires_identity
def list_projects_view(user, cache):
    """List cached projects."""
    projects = cache.get_projects(user)
    projects = [
        ProjectCloneResponse().load(p, unknown=EXCLUDE)
        for p in projects if make_project_path(user, p).exists()
    ]

    response = ProjectListResponseRPC().load({
        'result': ProjectListResponse().load({'projects': projects})
    })
    return jsonify(response)



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Cleanup jobs."""
import json
import os
import shutil
import time

from renku.service.cache import FileManagementCache, ProjectManagementCache
from renku.service.utils import make_file_path, make_project_path


def cache_files_cleanup():
    """Cache files a cleanup job."""
    ttl = int(os.getenv('RENKU_SVC_CLEANUP_TTL_FILES', 1800))
    cache = FileManagementCache()

    for file_key in cache.all_files_iter():
        try:
            user, _ = file_key.decode('utf-8').split('_')
        except ValueError:
            continue

        user = {'user_id': user}
        for file_meta in cache.hash_table(file_key).values():
            data = json.loads(file_meta)

            # If record does not contain timestamp,
            # it means its an old record, and we should mark it for deletion.
            created_at = data.get('timestamp', (time.time() - ttl) * 1e+3)
            old = ((time.time() * 1e+3) - created_at) / 1e+3

            if old >= ttl:
                file_path = make_file_path(user, data)
                if file_path.exists() and file_path.is_file():
                    try:
                        file_path.unlink()
                    except FileNotFoundError:
                        pass

                if file_path.exists() and file_path.is_dir():
                    shutil.rmtree(str(file_path))

                cache.invalidate_file(user, data['file_id'])


def cache_project_cleanup():
    """Cache project a cleanup job."""
    ttl = int(os.getenv('RENKU_SVC_CLEANUP_TTL_PROJECTS', 1800))
    cache = ProjectManagementCache()

    for file_key in cache.all_projects_iter():
        try:
            user, _ = file_key.decode('utf-8').split('_')
        except ValueError:
            continue

        user = {'user_id': user}
        for file_meta in cache.hash_table(file_key).values():
            data = json.loads(file_meta)

            # If record does not contain timestamp,
            # it means its an old record, and we should mark it for deletion.
            created_at = data.get('timestamp', (time.time() - ttl) * 1e+3)
            old = ((time.time() * 1e+3) - created_at) / 1e+3

            if old >= ttl:
                project_path = make_project_path(user, data)
                if project_path.exists():
                    shutil.rmtree(str(project_path))

                cache.invalidate_project(user, data['project_id'])



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test execution of commands in parallel."""

import subprocess
import sys
import time
from pathlib import Path


def test_run_in_isolation(runner, project, client, run):
    """Test run in isolation."""
    import filelock

    cwd = Path(project)
    with client.commit():
        with (cwd / '.gitignore').open('a') as f:
            f.write('lock')

    prefix = [
        'run',
        '--no-output',
    ]
    cmd = [
        'python', '-S', '-c',
        'import os, sys; sys.exit(1 if os.path.exists("lock") else 0)'
    ]

    head = client.repo.head.commit.hexsha

    with filelock.FileLock('lock'):
        assert 1 == run(args=prefix + cmd)
        assert client.repo.head.commit.hexsha == head

        assert 0 == run(prefix + ['--isolation'] + cmd)
        assert client.repo.head.commit.hexsha != head


def test_file_modification_during_run(tmpdir, runner, project, client, run):
    """Test run in isolation."""
    script = client.path / 'script.py'
    output = client.path / 'output'
    lock = Path(str(tmpdir.join('lock')))

    with client.commit():
        with script.open('w') as fp:
            fp.write(
                'import os, time, sys\n'
                'open("{lock}", "a")\n'
                'while os.path.exists("{lock}"):\n'
                '    time.sleep(1)\n'
                'sys.stdout.write(sys.stdin.read())\n'
                'sys.stdout.flush()\n'.format(lock=str(lock))
            )

    prefix = [
        sys.executable,
        '-m',
        'renku.cli',
        'run',
        '--isolation',
    ]
    cmd = [
        'python',
        script.name,
    ]

    previous = client.repo.head.commit

    with output.open('wb') as stdout:
        process = subprocess.Popen(
            prefix + cmd, stdin=subprocess.PIPE, stdout=stdout
        )

        while not lock.exists():
            time.sleep(1)

        with script.open('w') as fp:
            fp.write('print("edited")')

        lock.unlink()

        process.communicate(input=b'test')
        assert 0 == process.wait()

    with output.open('r') as fp:
        assert 'test' == fp.read().strip()

    diff = previous.diff(client.repo.head.commit)
    modifications = [
        modification
        for modification in diff if modification.change_type == 'M'
    ]
    assert 0 == len(modifications)



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Migrations for dataset."""
import os
import uuid
from pathlib import Path


def migrate_dataset_schema(data, client):
    """Migrate from old dataset formats."""
    if 'authors' not in data:
        return

    data['@context']['creator'] = data['@context'].pop(
        'authors', {'@container': 'list'}
    )

    data['creator'] = data.pop('authors', {})

    files = data.get('files', [])

    if isinstance(files, dict):
        files = files.values()
    for file_ in files:
        file_['creator'] = file_.pop('authors', {})

    return data


def migrate_absolute_paths(data, client):
    """Migrate dataset paths to use relative path."""
    raw_path = data.get('path', '.')
    path = Path(raw_path)

    if path.is_absolute():
        try:
            data['path'] = path.relative_to(os.getcwd())
        except ValueError:
            elements = raw_path.split('/')
            index = elements.index('.renku')
            data['path'] = Path('/'.join(elements[index:]))

    files = data.get('files', [])

    if isinstance(files, dict):
        files = list(files.values())

    for file_ in files:
        path = Path(file_.get('path'), '.')
        if path.is_absolute():
            file_['path'] = path.relative_to((os.getcwd()))
    data['files'] = files
    return data


def migrate_doi_identifier(data, client):
    """If the dataset _id is doi, make it a UUID."""
    from renku.core.utils.doi import is_doi
    from renku.core.utils.uuid import is_uuid

    _id = data.get('_id', '')
    identifier = data.get('identifier', '')

    if not is_uuid(_id):
        if not is_uuid(identifier):
            data['identifier'] = str(uuid.uuid4())
        if is_doi(data.get('_id', '')):
            data['same_as'] = {'@type': ['schema:URL'], 'url': data['_id']}
            if data.get('@context'):
                data['@context'].setdefault(
                    'same_as', {
                        '@id': 'schema:sameAs',
                        '@type': 'schema:URL',
                        '@context': {
                            '@version': '1.1',
                            'url': 'schema:url',
                            'schema': 'http://schema.org/'
                        }
                    }
                )
        data['_id'] = data['identifier']
    return data


def migrate_same_as_structure(data, client):
    """Changes sameAs string to schema:URL object."""
    same_as = data.get('same_as')

    if same_as and isinstance(same_as, str):
        data['same_as'] = {'@type': ['schema:URL'], 'url': same_as}

        if data.get('@context'):
            data['@context'].setdefault(
                'same_as', {
                    '@id': 'schema:sameAs',
                    '@type': 'schema:URL',
                    '@context': {
                        '@version': '1.1',
                        'url': 'schema:url',
                        'schema': 'http://schema.org/'
                    }
                }
            )

    return data


def migrate_dataset_file_id(data, client):
    """Ensure dataset files have a fully qualified url as id."""
    host = 'localhost'
    if client:
        host = client.remote.get('host') or host
    host = os.environ.get('RENKU_DOMAIN') or host

    files = data.get('files', [])
    for file_ in files:
        if not file_['_id'].startswith('http'):
            file_['_id'] = 'https://{host}/{id}'.format(
                host=host, id=file_['_id']
            )

    return data



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compatibility layer for different Python versions."""

import cgi
import contextlib
import html
import json
import os
import sys
from pathlib import Path

import pyld

if sys.version_info < (3, 6):
    original_resolve = Path.resolve

    def resolve(self, strict=False):
        """Support strict parameter."""
        if strict:
            return original_resolve(self)
        return Path(os.path.realpath(os.path.abspath(str(self))))

    Path.resolve = resolve

try:
    contextlib.nullcontext
except AttributeError:

    class nullcontext(object):
        """Context manager that does no additional processing.

        Used as a stand-in for a normal context manager, when a particular
        block of code is only sometimes used with a normal context manager:
        cm = optional_cm if condition else nullcontext()
        with cm:
            # Perform operation, using optional_cm if condition is True
        """

        def __init__(self, enter_result=None):
            self.enter_result = enter_result

        def __enter__(self):
            return self.enter_result

        def __exit__(self, *excinfo):
            pass

    contextlib.nullcontext = nullcontext

try:
    FileNotFoundError
except NameError:  # pragma: no cover
    FileNotFoundError = IOError


class PatchedActiveContextCache(pyld.jsonld.ActiveContextCache):
    """Pyld context cache without issue of missing contexts."""

    def set(self, active_ctx, local_ctx, result):
        if len(self.order) == self.size:
            entry = self.order.popleft()
            if sum(
                e['activeCtx'] == entry['activeCtx'] and
                e['localCtx'] == entry['localCtx'] for e in self.order
            ) == 0:
                # only delete from cache if it doesn't exist in context deque
                del self.cache[entry['activeCtx']][entry['localCtx']]
        key1 = json.dumps(active_ctx)
        key2 = json.dumps(local_ctx)
        self.order.append({'activeCtx': key1, 'localCtx': key2})
        self.cache.setdefault(key1, {})[key2] = json.loads(json.dumps(result))


pyld.jsonld._cache = {'activeCtx': PatchedActiveContextCache()}

cgi.escape = html.escape

__all__ = (
    'FileNotFoundError',
    'Path',
    'contextlib',
    'pyld',
    'cgi',
)



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service datasets view."""
import json
import os
import uuid
from pathlib import Path

from flask import Blueprint, jsonify, request
from flask_apispec import marshal_with, use_kwargs
from marshmallow import EXCLUDE

from renku.core.commands.dataset import add_file, create_dataset, \
    list_datasets, list_files
from renku.core.utils.contexts import chdir
from renku.service.config import INTERNAL_FAILURE_ERROR_CODE, \
    INVALID_PARAMS_ERROR_CODE, SERVICE_PREFIX
from renku.service.jobs.constants import USER_JOB_STATE_ENQUEUED
from renku.service.jobs.contexts import enqueue_retry
from renku.service.jobs.datasets import dataset_import
from renku.service.jobs.queues import DATASETS_JOB_QUEUE
from renku.service.serializers.datasets import DatasetAddRequest, \
    DatasetAddResponse, DatasetAddResponseRPC, DatasetCreateRequest, \
    DatasetCreateResponse, DatasetCreateResponseRPC, DatasetDetails, \
    DatasetFileDetails, DatasetFilesListRequest, DatasetFilesListResponse, \
    DatasetFilesListResponseRPC, DatasetImportRequest, \
    DatasetImportResponseRPC, DatasetListRequest, DatasetListResponse, \
    DatasetListResponseRPC
from renku.service.utils import make_file_path, make_project_path, repo_sync
from renku.service.views.decorators import accepts_json, handle_base_except, \
    handle_git_except, handle_renku_except, handle_validation_except, \
    header_doc, requires_cache, requires_identity

DATASET_BLUEPRINT_TAG = 'datasets'
dataset_blueprint = Blueprint(
    DATASET_BLUEPRINT_TAG, __name__, url_prefix=SERVICE_PREFIX
)


@use_kwargs(DatasetListRequest, locations=['query'])
@marshal_with(DatasetListResponseRPC)
@header_doc('List all datasets in project.', tags=(DATASET_BLUEPRINT_TAG, ))
@dataset_blueprint.route(
    '/datasets.list',
    methods=['GET'],
    provide_automatic_options=False,
)
@handle_base_except
@handle_git_except
@handle_renku_except
@handle_validation_except
@requires_cache
@requires_identity
def list_datasets_view(user, cache):
    """List all datasets in project."""
    req = DatasetListRequest().load(request.args)
    project = cache.get_project(user, req['project_id'])
    project_path = make_project_path(user, project)

    if not project_path:
        return jsonify(
            error={
                'code': INVALID_PARAMS_ERROR_CODE,
                'reason': 'invalid project_id argument',
            }
        )

    with chdir(project_path):
        datasets = [
            DatasetDetails().load(ds, unknown=EXCLUDE)
            # TODO: fix core interface to address this issue: #1022
            for ds in json.loads(list_datasets(None, 'data', 'json-ld'))
        ]

    response = DatasetListResponse().load({'datasets': datasets})
    return jsonify(DatasetListResponseRPC().load({'result': response}))


@use_kwargs(DatasetFilesListRequest, locations=['query'])
@marshal_with(DatasetFilesListResponseRPC)
@header_doc('List files in a dataset.', tags=(DATASET_BLUEPRINT_TAG, ))
@dataset_blueprint.route(
    '/datasets.files_list',
    methods=['GET'],
    provide_automatic_options=False,
)
@handle_base_except
@handle_git_except
@handle_renku_except
@handle_validation_except
@requires_cache
@requires_identity
def list_dataset_files_view(user, cache):
    """List files in a dataset."""
    ctx = DatasetFilesListRequest().load(request.args)
    project = cache.get_project(user, ctx['project_id'])
    project_path = make_project_path(user, project)

    if not project_path:
        return jsonify(
            error={
                'code': INVALID_PARAMS_ERROR_CODE,
                'reason': 'invalid project_id argument',
            }
        )

    with chdir(project_path):
        dataset_files = json.loads(
            # TODO: fix core interface to address this issue (add ticket ref)
            list_files(ctx['dataset_name'], None, None, None, 'json-ld')
        )
        ctx['files'] = [
            DatasetFileDetails().load(ds, unknown=EXCLUDE)
            for ds in dataset_files
        ]

    response = DatasetFilesListResponse().load(ctx, unknown=EXCLUDE)
    return jsonify(DatasetFilesListResponseRPC().load({'result': response}))


@use_kwargs(DatasetAddRequest)
@marshal_with(DatasetAddResponseRPC)
@header_doc(
    'Add an uploaded file to a cloned repository.',
    tags=(DATASET_BLUEPRINT_TAG, )
)
@dataset_blueprint.route(
    '/datasets.add',
    methods=['POST'],
    provide_automatic_options=False,
)
@handle_base_except
@handle_git_except
@handle_renku_except
@handle_validation_except
@accepts_json
@requires_cache
@requires_identity
def add_file_to_dataset_view(user, cache):
    """Add the uploaded file to cloned repository."""
    ctx = DatasetAddRequest().load(request.json)
    project = cache.get_project(user, ctx['project_id'])
    project_path = make_project_path(user, project)

    if not project_path:
        return jsonify(
            error={
                'code': INVALID_PARAMS_ERROR_CODE,
                'message': 'invalid project_id: {0}'.format(ctx['project_id']),
            }
        )

    if not ctx['commit_message']:
        ctx['commit_message'] = 'service: dataset add {0}'.format(
            ctx['dataset_name']
        )

    local_paths = []
    for _file in ctx['files']:
        local_path = None

        if 'file_id' in _file:
            file = cache.get_file(user, _file['file_id'])
            local_path = make_file_path(user, file)

        elif 'file_path' in _file:
            local_path = project_path / Path(_file['file_path'])

        if not local_path or not local_path.exists():
            return jsonify(
                error={
                    'code': INVALID_PARAMS_ERROR_CODE,
                    'message':
                        'invalid file reference: {0}'.
                        format(local_path.relative_to(project_path))
                }
            )

        ctx['commit_message'] += ' {0}'.format(local_path.name)
        local_paths.append(str(local_path))

    with chdir(project_path):
        add_file(
            local_paths,
            ctx['dataset_name'],
            create=ctx['create_dataset'],
            commit_message=ctx['commit_message']
        )

        if not repo_sync(project_path):
            return jsonify(
                error={
                    'code': INTERNAL_FAILURE_ERROR_CODE,
                    'message': 'repo sync failed'
                }
            )

    return jsonify(
        DatasetAddResponseRPC().load({
            'result': DatasetAddResponse().load(ctx, unknown=EXCLUDE)
        })
    )


@use_kwargs(DatasetCreateRequest)
@marshal_with(DatasetCreateResponseRPC)
@header_doc(
    'Create a new dataset in a project.', tags=(DATASET_BLUEPRINT_TAG, )
)
@dataset_blueprint.route(
    '/datasets.create',
    methods=['POST'],
    provide_automatic_options=False,
)
@handle_base_except
@handle_git_except
@handle_renku_except
@handle_validation_except
@accepts_json
@requires_cache
@requires_identity
def create_dataset_view(user, cache):
    """Create a new dataset in a project."""
    ctx = DatasetCreateRequest().load(request.json)
    project = cache.get_project(user, ctx['project_id'])

    project_path = make_project_path(user, project)
    if not project_path:
        return jsonify(
            error={
                'code': INVALID_PARAMS_ERROR_CODE,
                'message': 'invalid project_id argument',
            }
        )

    with chdir(project_path):
        create_dataset(
            ctx['dataset_name'],
            commit_message=ctx['commit_message'],
            creators=ctx.get('creators'),
            description=ctx.get('description'),
        )

    if not repo_sync(project_path):
        return jsonify(
            error={
                'code': INTERNAL_FAILURE_ERROR_CODE,
                'reason': 'push to remote failed silently - try again'
            }
        )

    return jsonify(
        DatasetCreateResponseRPC().load({
            'result': DatasetCreateResponse().load(ctx, unknown=EXCLUDE)
        })
    )


@use_kwargs(DatasetImportRequest)
@marshal_with(DatasetImportResponseRPC)
@header_doc('Import a dataset', tags=(DATASET_BLUEPRINT_TAG, ))
@dataset_blueprint.route(
    '/datasets.import',
    methods=['POST'],
    provide_automatic_options=False,
)
@handle_base_except
@handle_validation_except
@requires_cache
@requires_identity
def import_dataset_view(user, cache):
    """Import a dataset view."""
    ctx = DatasetImportRequest().load(request.json)
    project = cache.get_project(user, ctx['project_id'])
    project_path = make_project_path(user, project)

    if not project_path:
        return jsonify(
            error={
                'code': INVALID_PARAMS_ERROR_CODE,
                'message': 'invalid project_id: {0}'.format(ctx['project_id']),
            }
        )

    user_job = {
        'job_id': uuid.uuid4().hex,
        'state': USER_JOB_STATE_ENQUEUED,
    }
    cache.create_job(user, user_job)

    with enqueue_retry(DATASETS_JOB_QUEUE) as queue:
        queue.enqueue(
            dataset_import,
            user,
            user_job['job_id'],
            ctx['project_id'],
            ctx['dataset_uri'],
            short_name=ctx.get('short_name'),
            extract=ctx.get('extract', False),
            timeout=int(os.getenv('WORKER_DATASET_JOBS_TIMEOUT', 1800)),
            result_ttl=int(os.getenv('WORKER_DATASET_JOBS_RESULT_TTL', 500))
        )

    return jsonify(DatasetImportResponseRPC().dump({'result': user_job}))
